# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_meta.ipynb.

# %% auto 0
__all__ = ['DataTypedEnumMeta', 'FallbackEnumMeta', 'TypedEnumMeta']

# %% ../nbs/01_meta.ipynb 4
from enum import EnumMeta, auto
from typing import Optional, Type, List, Tuple, Any, Union, Dict

# %% ../nbs/01_meta.ipynb 6
class DataTypedEnumMeta(EnumMeta):
    '''Dataclass-like type checking for Enums

    Parameters
    ----------
    dtype : float, optional
        The data-type of the enum.
    '''

    # Dataclass-like type checking
    dtype: Optional[Type] = None

# %% ../nbs/01_meta.ipynb 8
class FallbackEnumMeta(EnumMeta):
    '''Fallback (default) member support for Enums

    Parameters
    ----------
    fbmem : any, optional
        The fallback member to use if provided a missing or invalid enum member

    force : bool, optional
        Whether or not to force fallback by using `cls.members` if fallback member is `None`
    '''
    # Fallback Member to use if provided a missing or invalid enum member
    fbmem: Optional[Any] = None

    # Whether or not to force fallback by using `cls.members` if fallback member is None
    force: Optional[bool] = False

# %% ../nbs/01_meta.ipynb 10
class TypedEnumMeta(DataTypedEnumMeta, FallbackEnumMeta):
    '''Typed enums with fallback members.

    Parameters
    ----------
    dtype : float, optional
        The data-type of the enum.

    fbmem : any, optional
        The fallback member to use if provided a missing or invalid enum member
    '''
    @classmethod
    def __prepare__(metacls, name, bases, **kwargs):
        '''Prepare the class namespace for the enum class

        Parameters
        ----------
        name : str
            Name of the enum class being prepared.
        bases : tuple
            Base classes of the enum class being prepared.
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        dict
            Dictionary for the class namespace.
        '''
        enum_dict = super().__prepare__(name, bases, **kwargs)
        # NOTE: this will through an error since we are using StrEnum
        # enum_dict['fbmem'] = None
        return enum_dict

    def __init__(self, clsname, bases, clsdict, **kwargs):
        '''Initialize the enum class

        Parameters
        ----------
        clsname : str
            Name of the enum class being initialized.
        bases : tuple
            Base classes of the enum class being initialized.
        clsdict : dict
            Dictionary of class attributes and methods.
        **kwargs : dict
            Additional keyword arguments.
        '''
        super().__init__(clsname, bases, clsdict, **kwargs)

    def __new__(
        metacls, cls, bases, clsdict, *, 
        dtype: Optional[Type] = None,
        fbmem: Optional[Union[str, Any]] = None,
        force: Optional[bool] = False,
        items: Optional[List[str]] = None,
    ):
        '''Create a new instance of the enum class

        Parameters
        ----------
        cls : str
            Name of the new class.
        bases : tuple
            Base classes of the new class.
        clsdict : dict
            Dictionary of class attributes and methods.
        dtype : type, optional
            Data type of the enum.
        fbmem : any, optional
            Fallback member of the enum.
        force: bool, default: False
            force failback by leveraging `cls.members[0]` if `fbmem` is not provided
        items : list of str, optional
            Enum members.

        Returns
        -------
        Enum
            The new enum class.
        '''
        if items is not None:
            for item in items:
                clsdict[item.upper()] = auto()

        new_cls = super().__new__(metacls, cls, bases, clsdict)

        # NOTE: this will result in TypeError: cannot extend 
        if fbmem:
            setattr(new_cls, 'fbmem', fbmem)

        if force:
            setattr(new_cls, 'force', force)

        if dtype:
            setattr(new_cls, 'dtype', dtype)
        
        return new_cls

    @property
    def members(cls) -> List[str]:
        '''List of names of the enum members.'''
        # NOTE: could also use cls._member_names_
        return [member.name for member in cls]

    @property
    def values(cls) -> List[Any]:
        '''List of values of the enum members.'''
        return [member.value for member in cls]
    
    @property
    def memdict(cls) -> Dict[str, Any]:
        '''Dict of (name, value) pairs of the enum members.'''
        return dict(zip(cls.members, cls.values))

    @property
    def items(cls) -> Dict[str, Any]:
        '''The items iterator for the dict of (name, value) pairs of the enum members.'''
        return cls.memdict.items()

    @property
    def tuples(cls) -> List[Tuple[str, Any]]:
        '''List of (name, value) pairs of the enum members.'''
        return list(zip(cls.members, cls.values))

    @property
    def clsname(cls) -> str:
        '''Name of the enum class.'''
        return cls.__name__
    
    @property
    def has_fbmem(cls) -> Optional[bool]:
        '''Check if the enum class has a fallback member.'''
        return hasattr(cls, 'fbmem') and getattr(cls, 'fbmem', None) is not None    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_decs.ipynb.

# %% auto 0
__all__ = ['set_fbmem', 'set_attr', 'set_default', 'set_fallback', 'set_method', 'set_dtype', 'set_repr']

# %% ../nbs/00_decs.ipynb 4
from enum import Enum, StrEnum, EnumMeta, auto
from typing import Optional, Any, Callable

# %% ../nbs/00_decs.ipynb 6
def set_attr(
    attr: str, 
    value: Optional[Any] = None,
    is_private: Optional[bool] = True
):
    """
    A decorator to set any attr of a class.

    Parameters
    ----------
    attr : str 
        The attribute to change value.

    value : Any, default: None
        The value to set the attribute to, by default None.

    is_private : bool, default: True
        Whether the attribute should be private or not, by default True.
        If True, the attribute will be prefixed with an underscore.
    
    Returns
    -------
    cls : type    
        The decorated class with the attribute set.
    """
    def wrapper(cls):
        prefix = '_' if is_private else ''
        attr_name = f'{prefix}{attr}'
        setattr(cls, attr_name, value)
        return cls
    return wrapper

# %% ../nbs/00_decs.ipynb 10
def set_default(default: str = ''):
    """
    A decorator to set the default member for a class.

    Parameters
    ----------    
    default : str, default: ''
        The default value for `_default`
    
    Returns
    -------
    cls : type    
        The decorated class with the `_default` attribute set.

    See Also
    --------
    set_attr : Sets an attribute for a class.
    """
    return set_attr(attr='default', value=default)

# %% ../nbs/00_decs.ipynb 13
def set_fallback(fbmem):
    """
    A decorator to set the default member for a class.

    Parameters
    ----------    
    fbmem : str
        The default fallback member or member name.
    
    Returns
    -------
    cls : type    
        The decorated class with the `fbmem` attribute set.

    See Also
    --------
    set_attr : Sets an attribute for a class.
    """
    return set_attr(attr='fbmem', value=fbmem, is_private=False)

set_fbmem = set_fallback

# %% ../nbs/00_decs.ipynb 16
def set_method(method):
    """
    A decorator to set the method attribute for a class.

    Parameters
    ----------    
    method : Callable
        The method to be set.
    
    Returns
    -------
    cls : type    
        The decorated class with the `method` attribute set.

    See Also
    --------
    set_attr : Sets an attribute for a class.
    """
    return set_attr(attr='method', value=method, is_private=False)

# %% ../nbs/00_decs.ipynb 19
def set_dtype(dtype: type):
    """
    A decorator to set the dtype for a class.

    Parameters
    ----------    
    dtype : type
        The dtype to be set.
    
    Returns
    -------
    cls : type    
        The decorated class with the `dtype` attribute set.

    See Also
    --------
    set_attr : Sets an attribute for a class.
    """    
    return set_attr(attr='dtype', value=dtype, is_private=False)

# %% ../nbs/00_decs.ipynb 22
def set_repr(repr_func: Callable):
    """
    A decorator to set the rich repr attributes (html & md) for a LiteralEnum class.

    Parameters
    ----------    
    repr_func : Callable
        The method to be set to `cls._repr_html_` and `cls._repr_markdown_`.
    
    Returns
    -------
    cls : type    
        The decorated class with the `_repr_html_` and `_repr_markdown_` attributes set.

    See Also
    --------
    set_attr : Sets an attribute for a class.
    """    
    def wrapper(cls):
        setattr(cls, '_repr_html_', repr_func)
        setattr(cls, '_repr_markdown_', repr_func)
        return cls
    return wrapper



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_enum.ipynb.

# %% auto 0
__all__ = ['Litum', 'Strum', 'Intum', 'TypedFallbackMixin', 'BaseTypedEnum', 'OperatorEnum', 'LiteralEnum', 'LitStrEnum',
           'LitIntEnum']

# %% ../nbs/02_enum.ipynb 4
from enum import Enum, IntEnum, StrEnum, auto
from typing import Optional, Callable, List, Tuple, Any, Dict, Union, Literal
from dataclasses import dataclass

# %% ../nbs/02_enum.ipynb 5
from .decs import set_attr, set_default, set_fallback, set_fbmem, set_method, set_dtype, set_repr
from .meta import DataTypedEnumMeta, FallbackEnumMeta, TypedEnumMeta

# %% ../nbs/02_enum.ipynb 7
class TypedFallbackMixin(metaclass=TypedEnumMeta):
    '''Mixin providing the fallback functionality for the TypedEnum class.'''

    @staticmethod
    def val_is_member_name(mem, val) -> bool:
        '''Check if the value matches the member name.

        Parameters
        ----------
        mem : EnumMember
            The enum member to check against.
        val : any
            The value to check.

        Returns
        -------
        bool
            True if the value matches the member name, False otherwise.
        '''
        return str(mem.name).lower() == str(val).lower()

    @staticmethod
    def val_is_member_value(mem, val) -> bool:
        '''Check if the value matches the member value.

        Parameters
        ----------
        mem : EnumMember
            The enum member to check against.
        val : any
            The value to check.

        Returns
        -------
        bool
            True if the value matches the member value, False otherwise.
        '''
        return str(mem.value).lower() == str(val).lower()

    @classmethod
    def val_matches_member(cls, mem, val) -> bool:
        '''Check if the value matches the member name or value.

        Parameters
        ----------
        mem : EnumMember
            The enum member to check against.
        val : any
            The value to check.

        Returns
        -------
        bool
            True if the value matches the member name or value, False otherwise.
        '''
        is_key = cls.val_is_member_name(mem, val)
        is_val = cls.val_is_member_value(mem, val)
        return is_key or is_val

    @classmethod
    def match_val_to_member(cls, val: Any) -> Optional[Any]:
        '''Find the enum member that matches the value.

        Parameters
        ----------
        val : any
            The value to match.

        Returns
        -------
        EnumMember or None
            The matching enum member or None if not found.
        '''
        if val is None: 
            return None            
        for mem in cls:
            if cls.val_matches_member(mem, val):
                return mem
        return None

    @classmethod
    def check_dtype(cls, mem) -> bool:
        '''Check if the enum member matches the enum's data type.

        Parameters
        ----------
        mem : EnumMember
            The enum member to check.

        Returns
        -------
        bool
            True if the member matches the data type, False otherwise.
        '''
        if not hasattr(cls, 'dtype'):
            return False

        dtype = cls.dtype
        if not isinstance(dtype, type):
            dtype = type(dtype)

        return isinstance(mem, dtype)

    @classmethod
    def check_fallback(cls, force: bool = False) -> Optional[Any]:
        '''Check for a fallback member or value.

        Parameters
        ----------
        force : bool, default: False
            If True, force the fallback to the first member if no fallback is found. Default is False.

        Returns
        -------
        EnumMember : any, default: None
            The fallback member or value, or None if not found.
        '''
        
        fback = None

        if hasattr(cls, 'fbmem'):
            fbmem = getattr(cls, 'fbmem', None)            
            if isinstance(fbmem, cls):
                fback = fbmem

            elif cls.check_dtype(fbmem):
                fback = fbmem
                try:
                    fback = cls.match_val_to_member(fback)
                    if fback is not None:
                        return fback
                except:
                    fback = None

            elif isinstance(fbmem, str):
                fback = getattr(cls, fbmem, None)

        if fback is None and force:
            forced_member = cls.members[0]
            return cls._missing_(forced_member)

        return fback

    @classmethod
    def _missing_(cls, val: Any) -> Optional[Any]:
        '''Handle missing enum members.

        Parameters
        ----------
        val : any
            The val to match.

        Returns
        -------
        EnumMember : any, default: None
            The matching or fallback member or val or None if not found.
        '''      
        found = cls.match_val_to_member(val)
        if found is not None:
            return found

        fback = cls.check_fallback(force=getattr(cls, 'force', True))
        return fback

# %% ../nbs/02_enum.ipynb 11
class BaseTypedEnum(TypedFallbackMixin, Enum):
    '''Base class for typed enums with fallback functionality.

    Examples
    --------
    >>> class MyEnum(BaseTypedEnum, dtype=int, fbmem='THREE'):
    ...    ONE = 1
    ...    TWO = 2
    ...    THREE = 3

    >>> print(MyEnum(1))
    MyEnum.ONE

    >>> print(MyEnum('missing'))
    MyEnum.THREE
    '''   
    pass

# %% ../nbs/02_enum.ipynb 16
class OperatorEnum(BaseTypedEnum):#, Enum, metaclass=TypedEnumMeta):
    '''Enum for operator functions.

    Examples
    --------
    >>> import operator
    >>> class MyOpEnum(OperatorEnum):
    ...    EQ = operator.eq
    ...    NE = operator.ne

    >>> eq = MyOpEnum('EQ')
    >>> print(eq(1, 1))
    True

    >>> ne = MyOpEnum('NE')
    >>> print(ne(1, 2))
    True
    '''    
    def __call__(self, *args, **kwargs) -> bool:
        '''Call the operator function.

        Parameters
        ----------
        *args
            Arguments to pass to the operator function.
        **kwargs
            Keyword arguments to pass to the operator function.

        Returns
        -------
        bool
            The result of the operator function.
        '''
        func = self.value
        return func(*args, **kwargs)

# %% ../nbs/02_enum.ipynb 21
class LiteralEnum(BaseTypedEnum, StrEnum, metaclass=TypedEnumMeta):
    '''Enum for stringly-typed enums like a keyword argument with a few valid str values.

    Examples
    --------
    >>> class MyLiteralEnum(LiteralEnum):
    ...    RED = 'red'
    ...    BLUE = 'blue'
    ...    GREEN = 'green'

    >>> color = MyLiteralEnum('RED')
    >>> print(color)
    MyLiteralEnum.RED

    >>> try:
    ...     invalid_color = MyLiteralEnum('ORANGE')
    ... except ValueError as e:
    ...     print(e)
    'ORANGE' is not a valid MyLiteralEnum
    '''
    pass

Litum = LiteralEnum

# %% ../nbs/02_enum.ipynb 28
class LitStrEnum(BaseTypedEnum, StrEnum, metaclass=TypedEnumMeta):    
    pass

Strum = LitStrEnum

# %% ../nbs/02_enum.ipynb 30
class LitIntEnum(BaseTypedEnum, IntEnum, metaclass=TypedEnumMeta):    
    pass

Intum = LitIntEnum



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/litum',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/litum',
                'lib_path': 'litum'},
  'syms': { 'litum.decs': { 'litum.decs.set_attr': ('decs.html#set_attr', 'litum/decs.py'),
                            'litum.decs.set_default': ('decs.html#set_default', 'litum/decs.py'),
                            'litum.decs.set_dtype': ('decs.html#set_dtype', 'litum/decs.py'),
                            'litum.decs.set_fallback': ('decs.html#set_fallback', 'litum/decs.py'),
                            'litum.decs.set_method': ('decs.html#set_method', 'litum/decs.py'),
                            'litum.decs.set_repr': ('decs.html#set_repr', 'litum/decs.py')},
            'litum.enum': { 'litum.enum.BaseTypedEnum': ('enum.html#basetypedenum', 'litum/enum.py'),
                            'litum.enum.LitIntEnum': ('enum.html#litintenum', 'litum/enum.py'),
                            'litum.enum.LitStrEnum': ('enum.html#litstrenum', 'litum/enum.py'),
                            'litum.enum.LiteralEnum': ('enum.html#literalenum', 'litum/enum.py'),
                            'litum.enum.OperatorEnum': ('enum.html#operatorenum', 'litum/enum.py'),
                            'litum.enum.OperatorEnum.__call__': ('enum.html#operatorenum.__call__', 'litum/enum.py'),
                            'litum.enum.TypedFallbackMixin': ('enum.html#typedfallbackmixin', 'litum/enum.py'),
                            'litum.enum.TypedFallbackMixin._missing_': ('enum.html#typedfallbackmixin._missing_', 'litum/enum.py'),
                            'litum.enum.TypedFallbackMixin.check_dtype': ('enum.html#typedfallbackmixin.check_dtype', 'litum/enum.py'),
                            'litum.enum.TypedFallbackMixin.check_fallback': ( 'enum.html#typedfallbackmixin.check_fallback',
                                                                              'litum/enum.py'),
                            'litum.enum.TypedFallbackMixin.match_val_to_member': ( 'enum.html#typedfallbackmixin.match_val_to_member',
                                                                                   'litum/enum.py'),
                            'litum.enum.TypedFallbackMixin.val_is_member_name': ( 'enum.html#typedfallbackmixin.val_is_member_name',
                                                                                  'litum/enum.py'),
                            'litum.enum.TypedFallbackMixin.val_is_member_value': ( 'enum.html#typedfallbackmixin.val_is_member_value',
                                                                                   'litum/enum.py'),
                            'litum.enum.TypedFallbackMixin.val_matches_member': ( 'enum.html#typedfallbackmixin.val_matches_member',
                                                                                  'litum/enum.py')},
            'litum.meta': { 'litum.meta.DataTypedEnumMeta': ('meta.html#datatypedenummeta', 'litum/meta.py'),
                            'litum.meta.FallbackEnumMeta': ('meta.html#fallbackenummeta', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta': ('meta.html#typedenummeta', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.__init__': ('meta.html#typedenummeta.__init__', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.__new__': ('meta.html#typedenummeta.__new__', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.__prepare__': ('meta.html#typedenummeta.__prepare__', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.clsname': ('meta.html#typedenummeta.clsname', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.has_fbmem': ('meta.html#typedenummeta.has_fbmem', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.items': ('meta.html#typedenummeta.items', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.members': ('meta.html#typedenummeta.members', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.memdict': ('meta.html#typedenummeta.memdict', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.tuples': ('meta.html#typedenummeta.tuples', 'litum/meta.py'),
                            'litum.meta.TypedEnumMeta.values': ('meta.html#typedenummeta.values', 'litum/meta.py')}}}



__version__ = "0.0.1"

from .decs import (
    set_attr, set_default, set_fallback, set_fbmem, 
    set_method, set_dtype, set_repr
)

from .meta import (
    DataTypedEnumMeta, FallbackEnumMeta, TypedEnumMeta
)

from .enum import (
    TypedFallbackMixin, BaseTypedEnum, OperatorEnum, LiteralEnum,
    LitStrEnum, LitIntEnum, Litum, Strum, Intum
)

__all__ = [
    'set_attr', 'set_default', 'set_fallback', 'set_fbmem',
    'set_method', 'set_dtype', 'set_repr',

    'DataTypedEnumMeta', 'FallbackEnumMeta', 'TypedEnumMeta',

    'TypedFallbackMixin', 'BaseTypedEnum', 'OperatorEnum', 'LiteralEnum',
    'LitStrEnum', 'LitIntEnum', 
    'Litum', 'Strum', 'Intum', 
]